@@ -1,6 +1,6 @@
 #ifndef lint
 static char Rcs_Id[] =
-    "$Id: correct.c,v 1.83 2015-02-08 00:35:41-08 geoff Exp $";
+    "$Id: correct.c,v 1.87 2021-01-09 13:12:35-08 geoff Exp $";
 #endif
 
 /*
@@ -48,6 +48,21 @@ static char Rcs_Id[] =
 
 /*
  * $Log: correct.c,v $
+ * Revision 1.87  2021-01-09 13:12:35-08  geoff
+ * Suppress warnings about empty printf format strings.
+ *
+ * Revision 1.86  2020-12-30 22:33:41-08  geoff
+ * Support SQUEEZE_STRINGS and RAW_DISPLAY.  Also fix a compiler warning
+ * and a minor bug in counting possibilities.
+ *
+ * Revision 1.85  2020-10-26 11:39:55-07  geoff
+ * Include header files needed for C99 function declarations.  This probably
+ * breaks backwards compatibility but I don't have time to research it and
+ * add appropriate config.X options.  Sigh.
+ *
+ * Revision 1.84  2020-04-28 17:55:52-07  geoff
+ * Don't segfault if user input starts with a null byte
+ *
  * Revision 1.83  2015-02-08 00:35:41-08  geoff
  * Fix a namespace collision with "getline".
  *
@@ -204,6 +219,7 @@ static char Rcs_Id[] =
  */
 
 #include <ctype.h>
+#include <fcntl.h>
 #include "config.h"
 #include "ispell.h"
 #include "proto.h"
@@ -272,15 +288,33 @@ void givehelp (interactive)
     else
 	helpout = stderr;
 
+#ifdef CORR_C_HELP_1
     (void) fprintf (helpout, CORR_C_HELP_1, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_1 */
+#ifdef CORR_C_HELP_2
     (void) fprintf (helpout, CORR_C_HELP_2, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_2 */
+#ifdef CORR_C_HELP_3
     (void) fprintf (helpout, CORR_C_HELP_3, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_3 */
+#ifdef CORR_C_HELP_4
     (void) fprintf (helpout, CORR_C_HELP_4, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_4 */
+#ifdef CORR_C_HELP_5
     (void) fprintf (helpout, CORR_C_HELP_5, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_5 */
+#ifdef CORR_C_HELP_6
     (void) fprintf (helpout, CORR_C_HELP_6, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_6 */
+#ifdef CORR_C_HELP_7
     (void) fprintf (helpout, CORR_C_HELP_7, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_7 */
+#ifdef CORR_C_HELP_8
     (void) fprintf (helpout, CORR_C_HELP_8, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_8 */
+#ifdef CORR_C_HELP_9
     (void) fprintf (helpout, CORR_C_HELP_9, MAYBE_CR (helpout));
+#endif /* CORR_C_HELP_9 */
 
     (void) fprintf (helpout, CORR_C_HELP_COMMANDS, MAYBE_CR (helpout),
       MAYBE_CR (helpout), MAYBE_CR (helpout));
@@ -474,7 +508,7 @@ checkagain:
 #endif /* BOTTOMCONTEXT */
 	if (i >= easypossibilities)
 	    (void) printf ("??: %s", possibilities[i]);
-	else if (easypossibilities >= 10  &&  i < 10)
+	else if (easypossibilities > 10  &&  i < 10)
 	    (void) printf ("0%d: %s", i, possibilities[i]);
 	else
 	    (void) printf ("%2d: %s", i, possibilities[i]);
@@ -626,7 +660,7 @@ checkagain:
 	    case '0': case '1': case '2': case '3': case '4':
 	    case '5': case '6': case '7': case '8': case '9':
 		i = c - '0';
-		if (easypossibilities >= 10)
+		if (easypossibilities > 10)
 		    {
 		    c = GETKEYSTROKE ();
 		    if (c >= '0'  &&  c <= '9')
@@ -752,8 +786,22 @@ static int show_char (cp, linew, output, maxw)
 	(*cp)++;
 	return 8 - (linew & 0x07);
 	}
+    if (!vflag  &&  chartypes[defstringgroup].options & RAW_DISPLAY)
+        {
+        if (output)
+            {
+            for (i = 0;  i < len;  i++)
+                putchar(*(*cp)++);
+            }
+        else
+            *cp += len;
+        if (chartypes[defstringgroup].options & SQUEEZE_STRINGS)
+            return 1;
+        else
+            return len;
+        }
     /*
-     * Character is non-printing, or it's ISO and vflag is set.  Display
+     * Character is non-printing, or it's printing but vflag is set.  Display
      * it in "cat -v" form.  For string characters, display every element
      * separately in that form.
      */
@@ -827,7 +875,7 @@ static void inserttoken (buf, start, tok, curchar, oktochange)
 	for (p = start;  p != *curchar;  p++)
 	    (void) putc (*p, logfile);
 	(void) putc (' ', logfile);
-	(void) fputs (tok, logfile);
+	(void) fputs ((char *) tok, logfile);
 	(void) putc ('\n', logfile);
 	(void) fflush (logfile);
 	}
@@ -1708,6 +1756,8 @@ void askmode ()
 	 * all that likely).
 	 */
 	bufsize = strlen ((char *) filteredbuf);
+	if (bufsize == 0)
+            continue;
 	hadnl = filteredbuf[bufsize - 1] == '\n';
 	if (bufsize == (sizeof filteredbuf) / 2 - 1)
 	    {
