@@ -1,6 +1,6 @@
 #ifndef lint
 static char Rcs_Id[] =
-    "$Id: ispell.c,v 1.152 2001/07/25 21:51:46 geoff Exp $";
+    "$Id: ispell.c,v 1.161 2005/05/25 14:13:53 geoff Exp $";
 #endif
 
 #define MAIN
@@ -10,7 +10,7 @@ static char Rcs_Id[] =
  *
  * Copyright (c), 1983, by Pace Willisson
  *
- * Copyright 1992, 1993, 1999, 2001, Geoff Kuenning, Claremont, CA
+ * Copyright 1992, 1993, 1999, 2001, 2005, Geoff Kuenning, Claremont, CA
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,16 +26,8 @@ static char Rcs_Id[] =
  *    such.  Binary redistributions based on modified source code
  *    must be clearly marked as modified versions in the documentation
  *    and/or other materials provided with the distribution.
- * 4. Any web site or other electronic service that offers ispell for
- *    download or other electronic transfer as a separate entity, in
- *    either source or binary form, must also include a prominent statement
- *    indicating that information about ispell can be obtained from the
- *    following Web site URL:
- *	http://fmg-www.cs.ucla.edu/geoff/ispell.html
- *    If the offering service supports hyperlinks, the aforementioned
- *    Web site must also be offered as a hyperlink.  Condition #4 does
- *    not apply if ispell is offered only as part of a larger, aggregated
- *    product such as a word processor or packaged operating system.
+ * 4. The code that causes the 'ispell -v' command to display a prominent
+ *    link to the official ispell Web site may not be removed.
  * 5. The name of Geoff Kuenning may not be used to endorse or promote
  *    products derived from this software without specific prior
  *    written permission.
@@ -55,6 +47,39 @@ static char Rcs_Id[] =
 
 /*
  * $Log: ispell.c,v $
+ * Revision 1.161  2005/05/25 14:13:53  geoff
+ * Report the value of EXEEXT in "ispell -vv".
+ *
+ * Revision 1.160  2005/04/28 14:46:51  geoff
+ * Open a correction log file in command mode.
+ *
+ * Revision 1.159  2005/04/28 00:26:06  geoff
+ * Don't print the count-file suffix, since it's no longer used.
+ *
+ * Revision 1.158  2005/04/20 23:16:32  geoff
+ * Rename some variables to make them more meaningful.
+ *
+ * Revision 1.157  2005/04/14 23:11:36  geoff
+ * Move initckch to makedent.c.
+ *
+ * Revision 1.156  2005/04/14 21:25:52  geoff
+ * Make DICTIONARYVAR configurable, add CHARSETVAR, and add both to ispell -vv.
+ *
+ * Revision 1.155  2005/04/14 15:19:37  geoff
+ * Make sure all current config.X variables are represented in ispell -vv
+ * (with the exception of a couple that are function macros), and clean
+ * up the sorting of the printout.
+ *
+ * Revision 1.154  2005/04/14 14:38:23  geoff
+ * Update license.  Incorporate Ed Avis's changes.  Add new config.X
+ * variables to -vv output.  For backwards compatibility, always put
+ * !NO8BIT in config.X output.  Add the -o switch.  Add -e5.  Be
+ * intelligent about inferring the output file mode if there is an
+ * external deformatter.
+ *
+ * Revision 1.153  2001/09/06 00:30:28  geoff
+ * Many changes from Eli Zaretskii to support DJGPP compilation.
+ *
  * Revision 1.152  2001/07/25 21:51:46  geoff
  * Minor license update.
  *
@@ -172,7 +197,7 @@ static char Rcs_Id[] =
  * Allow the -t, -n, and -T switches to override each other, as follows:
  * if no switches are given, the deformatter and string characters are
  * chosen based on the file suffix.  If only -t/-n are given, the
- * deformatter is forced but string cahracters come from the file suffix.
+ * deformatter is forced but string characters come from the file suffix.
  * If only -T is given, the deformatter is chosen based on the value
  * given in the -T switch.  Finally, if both -T and -t/-n are given,
  * string characters are controlled by -T and the deformatter by -t/-n.
@@ -201,6 +226,7 @@ static char Rcs_Id[] =
 
 #include "config.h"
 #include "ispell.h"
+#include "fields.h"
 #include "proto.h"
 #include "msgs.h"
 #include "version.h"
@@ -211,12 +237,12 @@ static char Rcs_Id[] =
 #include <sys/stat.h>
 
 static void	usage P ((void));
-static void	initckch P ((char * wchars));
 int		main P ((int argc, char * argv[]));
 static void	dofile P ((char * filename));
-static FILE *	setupdefmt P ((char * filename));
+static FILE *	setupdefmt P ((char * filename, struct stat * statbuf));
 static void	update_file P ((char * filename, struct stat * statbuf));
 static void	expandmode P ((int printorig));
+char *		last_slash P ((char * file));
 
 static char *	Cmd;
 static char *	LibDict = NULL;		/* Pointer to name of $(LIBDIR)/dict */
@@ -235,99 +261,19 @@ static void usage ()
     exit (1);
     }
 
-static void initckch (wchars)
-    char *		wchars;		/* Characters in -w option, if any */
-    {
-    register ichar_t	c;
-    char		num[4];
-
-    for (c = 0; c < (ichar_t) (SET_SIZE + hashheader.nstrchars); ++c)
-	{
-	if (iswordch (c))
-	    {
-	    if (!mylower (c))
-		{
-		Try[Trynum] = c;
-		++Trynum;
-		}
-	    }
-	else if (isboundarych (c))
-	    {
-	    Try[Trynum] = c;
-	    ++Trynum;
-	    }
-	}
-    if (wchars != NULL)
-	{
-	while (Trynum < SET_SIZE + MAXSTRINGCHARS  &&  *wchars != '\0')
-	    {
-	    if (*wchars != 'n'  &&  *wchars != '\\')
-		{
-		c = *wchars;
-		++wchars;
-		}
-	    else
-		{
-		++wchars;
-		num[0] = '\0'; 
-		num[1] = '\0'; 
-		num[2] = '\0'; 
-		num[3] = '\0';
-		if (isdigit (wchars[0]))
-		    {
-		    num[0] = wchars[0];
-		    if (isdigit (wchars[1]))
-			{
-			num[1] = wchars[1];
-			if (isdigit (wchars[2]))
-			    num[2] = wchars[2];
-			}
-		    }
-		if (wchars[-1] == 'n')
-		    {
-		    wchars += strlen (num);
-		    c = atoi (num);
-		    }
-		else
-		    {
-		    wchars += strlen (num);
-		    c = 0;
-		    if (num[0])
-			c = num[0] - '0';
-		    if (num[1])
-			{
-			c <<= 3;
-			c += num[1] - '0';
-			}
-		    if (num[2])
-			{
-			c <<= 3;
-			c += num[2] - '0';
-			}
-		    }
-		}
-	    c &= 0xFF;
-	    if (!hashheader.wordchars[c])
-		{
-		hashheader.wordchars[c] = 1;
-		hashheader.sortorder[c] = hashheader.sortval++;
-		Try[Trynum] = c;
-		++Trynum;
-		}
-	    }
-	}
-    }
-
 int main (argc, argv)
     int		argc;
     char *	argv[];
     {
     char *	p;
+    char	libdir[MAXPATHLEN];
     char *	cpd;
+    field_t *	extra_args;	/* Extra arguments from OPTIONVAR */
     char **	versionp;
     char *	wchars = NULL;
     char *	preftype = NULL;
     static char	libdictname[sizeof DEFHASH];
+    char	logfilename[MAXPATHLEN];
     static char	outbuf[BUFSIZ];
     int		argno;
     int		arglen;
@@ -336,31 +282,80 @@ int main (argc, argv)
 
     Trynum = 0;
 
-    p = getenv ("DICTIONARY");
+    p = getenv (LIBRARYVAR);
+    if (p == NULL)
+	(void) strcpy (libdir, LIBDIR);
+    else
+	{
+	(void) strncpy (libdir, p, sizeof libdir);
+	libdir[sizeof libdir - 1] = '\0';
+	}
+
+    p = getenv (DICTIONARYVAR);
     if (p != NULL)
 	{
-	if (index (p, '/') != NULL)
+	if (last_slash (p) != NULL)
 	    (void) strcpy (hashname, p);
 	else
-	    (void) sprintf (hashname, "%s/%s", LIBDIR, p);
+	    (void) sprintf (hashname, "%s/%s", libdir, p);
 	(void) strcpy (libdictname, p);
 	p = rindex (p, '.');
 	if (p == NULL  ||  strcmp (p, HASHSUFFIX) != 0)
 	    (void) strcat (hashname, HASHSUFFIX);
-	LibDict = rindex (libdictname, '/');
+	LibDict = last_slash (libdictname);
 	if (LibDict != NULL)
 	    LibDict++;
 	else
 	    LibDict = libdictname;
-	p = rindex (libdictname, '.');
+	p = rindex (LibDict, '.');
 	if (p != NULL)
 	    *p = '\0';
 	}
-    else
-	(void) sprintf (hashname, "%s/%s", LIBDIR, DEFHASH);
+   else
+	(void) sprintf (hashname, "%s/%s", libdir, DEFHASH);
 
     cpd = NULL;
 
+    /*
+    ** If any options were given in OPTIONVAR, prepend them to the
+    ** command-line arguments.  We prepend so that the command-line
+    ** arguments can override those from the environment.
+    */
+    p = getenv (OPTIONVAR);
+    if (p != NULL)
+	{
+	char **		newargv;
+
+	extra_args = fieldmake (p, 0, " \t",
+	  FLD_RUNS | FLD_SNGLQUOTES | FLD_DBLQUOTES | FLD_SHQUOTES
+	    | FLD_STRIPQUOTES | FLD_BACKSLASH,
+	  0);
+	if (extra_args == NULL)
+	    {
+	    (void) fprintf (stderr, ISPELL_C_NO_OPTIONS_SPACE);
+	    return 1;
+	    }
+	else
+	    {
+	    newargv =
+	      (char **) calloc (argc + extra_args->nfields, sizeof (char *));
+	    if (newargv == NULL)
+		{
+		(void) fprintf (stderr, ISPELL_C_NO_OPTIONS_SPACE);
+		return 1;
+		}
+
+	    /* Copy arguments over */
+	    newargv[0] = argv[0];
+	    for (argno = 0;  argno < (int) extra_args->nfields;  argno++)
+		newargv[argno + 1] = extra_args->fields[argno];
+	    for (argc += extra_args->nfields, argno++;  argno < argc;  argno++)
+		newargv[argno] = argv[argno - extra_args->nfields];
+	    }
+
+	argv = newargv;
+	}
+
     argv++;
     argc--;
     while (argc && **argv == '-')
@@ -392,6 +387,10 @@ int main (argc, argv)
 		if ((*argv)[2] == 'v')
 		    {
 		    (void) printf (ISPELL_C_OPTIONS_ARE);
+		    /*
+		     * We print USG first because it's mis-set so often.
+		     * All others are in alphabetical order.
+		     */
 #ifdef USG
 		    (void) printf ("\tUSG\n");
 #else /* USG */
@@ -404,23 +403,22 @@ int main (argc, argv)
 #else /* BOTTOMCONTEXT */
 		    (void) printf ("\t!BOTTOMCONTEXT\n");
 #endif /* BOTTOMCONTEXT */
-#if TERM_MODE == CBREAK
-		    (void) printf ("\tCBREAK\n");
-#endif /* TERM_MODE */
 		    (void) printf ("\tCC = \"%s\"\n", CC);
 		    (void) printf ("\tCFLAGS = \"%s\"\n", CFLAGS);
+		    (void) printf ("\tCHARSETVAR = \"%s\"\n", CHARSETVAR);
 #ifdef COMMANDFORSPACE
 		    (void) printf ("\tCOMMANDFORSPACE\n");
 #else /* COMMANDFORSPACE */
 		    (void) printf ("\t!COMMANDFORSPACE\n");
 #endif /* COMMANDFORSPACE */
+		    (void) printf ("\tCONTEXTPCT = %d\n", CONTEXTPCT);
 #ifdef CONTEXTROUNDUP
 		    (void) printf ("\tCONTEXTROUNDUP\n");
 #else /* CONTEXTROUNDUP */
 		    (void) printf ("\t!CONTEXTROUNDUP\n");
 #endif /* CONTEXTROUNDUP */
-		    (void) printf ("\tCONTEXTPCT = %d\n", CONTEXTPCT);
-		    (void) printf ("\tCOUNTSUFFIX = \"%s\"\n", COUNTSUFFIX);
+		    (void) printf ("\tDEFAULT_FILE_MODE = 0%3.3o\n",
+		      DEFAULT_FILE_MODE);
 		    (void) printf ("\tDEFHASH = \"%s\"\n", DEFHASH);
 		    (void) printf ("\tDEFINCSTR = \"%s\"\n", DEFINCSTR);
 		    (void) printf ("\tDEFLANG = \"%s\"\n", DEFLANG);
@@ -429,24 +427,32 @@ int main (argc, argv)
 		    (void) printf ("\tDEFPAFF = \"%s\"\n", DEFPAFF);
 		    (void) printf ("\tDEFPDICT = \"%s\"\n", DEFPDICT);
 		    (void) printf ("\tDEFTEXFLAG = %d\n", DEFTEXFLAG);
+		    (void) printf ("\tDICTIONARYVAR = \"%s\"\n",
+		      DICTIONARYVAR);
 		    (void) printf ("\tEGREPCMD = \"%s\"\n", EGREPCMD);
 #ifdef EQUAL_COLUMNS
 		    (void) printf ("\tEQUAL_COLUMNS\n");
 #else /* EQUAL_COLUMNS */
 		    (void) printf ("\t!EQUAL_COLUMNS\n");
 #endif /* EQUAL_COLUMNS */
+		    (void) printf ("\tEXEEXT = \"%s\"\n", EXEEXT);
 #ifdef GENERATE_LIBRARY_PROTOS
 		    (void) printf ("\tGENERATE_LIBRARY_PROTOS\n");
 #else /* GENERATE_LIBRARY_PROTOS */
 		    (void) printf ("\t!GENERATE_LIBRARY_PROTOS\n");
 #endif /* GENERATE_LIBRARY_PROTOS */
+		    (void) printf ("\tHASHSUFFIX = \"%s\"\n", HASHSUFFIX);
 #ifdef HAS_RENAME
 		    (void) printf ("\tHAS_RENAME\n");
 #else /* HAS_RENAME */
 		    (void) printf ("\t!HAS_RENAME\n");
 #endif /* HAS_RENAME */
-		    (void) printf ("\tHASHSUFFIX = \"%s\"\n", HASHSUFFIX);
 		    (void) printf ("\tHOME = \"%s\"\n", HOME);
+		    (void) printf ("\tHTMLCHECK = \"%s\"\n", HTMLCHECK);
+		    (void) printf ("\tHTMLCHECKVAR = \"%s\"\n", HTMLCHECKVAR);
+		    (void) printf ("\tHTMLIGNORE = \"%s\"\n", HTMLIGNORE);
+		    (void) printf ("\tHTMLIGNOREVAR = \"%s\"\n",
+		      HTMLIGNOREVAR);
 #ifdef IGNOREBIB
 		    (void) printf ("\tIGNOREBIB\n");
 #else /* IGNOREBIB */
@@ -454,27 +460,38 @@ int main (argc, argv)
 #endif /* IGNOREBIB */
 		    (void) printf ("\tINCSTRVAR = \"%s\"\n", INCSTRVAR);
 		    (void) printf ("\tINPUTWORDLEN = %d\n", INPUTWORDLEN);
+		    (void) printf ("\tINSTALL = \"%s\"\n", INSTALL);
 		    (void) printf ("\tLANGUAGES = \"%s\"\n", LANGUAGES);
 		    (void) printf ("\tLIBDIR = \"%s\"\n", LIBDIR);
 		    (void) printf ("\tLIBES = \"%s\"\n", LIBES);
+		    (void) printf ("\tLIBRARYVAR = \"%s\"\n", LIBRARYVAR);
+		    (void) printf ("\tLINK = \"%s\"\n",  LINK);
 		    (void) printf ("\tLINT = \"%s\"\n", LINT);
 		    (void) printf ("\tLINTFLAGS = \"%s\"\n", LINTFLAGS);
 #ifndef REGEX_LOOKUP
 		    (void) printf ("\tLOOK = \"%s\"\n", LOOK);
 #endif /* REGEX_LOOKUP */
+		    (void) printf ("\tLOOK_XREF = \"%s\"\n", LOOK_XREF);
 		    (void) printf ("\tMAKE_SORTTMP = \"%s\"\n", MAKE_SORTTMP);
 		    (void) printf ("\tMALLOC_INCREMENT = %d\n",
 		      MALLOC_INCREMENT);
 		    (void) printf ("\tMAN1DIR = \"%s\"\n", MAN1DIR);
 		    (void) printf ("\tMAN1EXT = \"%s\"\n", MAN1EXT);
-		    (void) printf ("\tMAN4DIR = \"%s\"\n", MAN4DIR);
-		    (void) printf ("\tMAN4EXT = \"%s\"\n", MAN4EXT);
+		    (void) printf ("\tMAN45DIR = \"%s\"\n", MAN45DIR);
+		    (void) printf ("\tMAN45EXT = \"%s\"\n", MAN45EXT);
+		    (void) printf ("\tMAN45SECT = \"%s\"\n", MAN45SECT);
 		    (void) printf ("\tMASKBITS = %d\n", MASKBITS);
-		    (void) printf ("\tMASKTYPE = %s\n", MASKTYPE_STRING);
+		    (void) printf ("\tMASKTYPE = \"%s\"\n", MASKTYPE_STRING);
 		    (void) printf ("\tMASKTYPE_WIDTH = %d\n", MASKTYPE_WIDTH);
 		    (void) printf ("\tMASTERHASH = \"%s\"\n", MASTERHASH);
 		    (void) printf ("\tMAXAFFIXLEN = %d\n", MAXAFFIXLEN);
+#ifdef MAXBASENAMELEN
+		    (void) printf ("\tMAXBASENAMELEN = %d\n", MAXBASENAMELEN);
+#endif
 		    (void) printf ("\tMAXCONTEXT = %d\n", MAXCONTEXT);
+#ifdef MAXEXTLEN
+		    (void) printf ("\tMAXEXTLEN = %d\n", MAXEXTLEN);
+#endif
 		    (void) printf ("\tMAXINCLUDEFILES = %d\n",
 		      MAXINCLUDEFILES);
 		    (void) printf ("\tMAXNAMLEN = %d\n", MAXNAMLEN);
@@ -495,9 +512,29 @@ int main (argc, argv)
 		    (void) printf ("\t!MINIMENU\n");
 #endif /* MINIMENU */
 		    (void) printf ("\tMINWORD = %d\n", MINWORD);
+#ifdef MSDOS
+		    (void) printf ("\tMSDOS\n");
+#else /* MSDOS */
+		    (void) printf ("\t!MSDOS\n");
+#endif /* MSDSO */
 		    (void) printf ("\tMSDOS_BINARY_OPEN = 0x%x\n",
 		      (unsigned int) MSDOS_BINARY_OPEN);
-		    (void) printf ("\tMSGLANG = %s\n", MSGLANG);
+		    (void) printf ("\tMSGLANG = \"%s\"\n", MSGLANG);
+		    /*
+		     * NO8BIT is an obsolete option, but some people depend
+		     * on it being in the ispell -vv output.
+		     */
+		    (void) printf ("\t!NO8BIT (8BIT)\n");
+#ifdef NO_FCNTL_H
+		    (void) printf ("\tNO_FCNTL_H\n");
+#else /* NO_FCNTL_H */
+		    (void) printf ("\t!NO_FCNTL_H (FCNTL_H)\n");
+#endif /* NO_FCNTL_H */
+#ifdef NO_MKSTEMP
+		    (void) printf ("\tNO_MKSTEMP\n");
+#else /* NO_MKSTEMP */
+		    (void) printf ("\t!NO_MKSTEMP (MKSTEMP)\n");
+#endif /* NO_STDLIB_H */
 #ifdef NO_STDLIB_H
 		    (void) printf ("\tNO_STDLIB_H\n");
 #else /* NO_STDLIB_H */
@@ -506,6 +543,7 @@ int main (argc, argv)
 		    (void) printf ("\tNRSPECIAL = \"%s\"\n", NRSPECIAL);
 		    (void) printf ("\tOLDPAFF = \"%s\"\n", OLDPAFF);
 		    (void) printf ("\tOLDPDICT = \"%s\"\n", OLDPDICT);
+		    (void) printf ("\tOPTIONVAR = \"%s\"\n", OPTIONVAR);
 #ifdef PDICTHOME
 		    (void) printf ("\tPDICTHOME = \"%s\"\n", PDICTHOME);
 #else /* PDICTHOME */
@@ -517,21 +555,33 @@ int main (argc, argv)
 #else /* PIECEMEAL_HASH_WRITES */
 		    (void) printf ("\t!PIECEMEAL_HASH_WRITES\n");
 #endif /* PIECEMEAL_HASH_WRITES */
-#if TERM_MODE != CBREAK
-		    (void) printf ("\tRAW\n");
-#endif /* TERM_MODE */
+		    (void) printf ("\tPOUNDBANG = \"%s\"\n", POUNDBANG);
 #ifdef REGEX_LOOKUP
 		    (void) printf ("\tREGEX_LOOKUP\n");
 #else /* REGEX_LOOKUP */
 		    (void) printf ("\t!REGEX_LOOKUP\n");
 #endif /* REGEX_LOOKUP */
 		    (void) printf ("\tREGLIB = \"%s\"\n", REGLIB);
-		    (void) printf ("\tSIGNAL_TYPE = %s\n", SIGNAL_TYPE_STRING);
+		    (void) printf ("\tR_OK = %d\n", R_OK);
+		    (void) printf ("\tSIGNAL_TYPE = \"%s\"\n",
+		      SIGNAL_TYPE_STRING);
 		    (void) printf ("\tSORTPERSONAL = %d\n", SORTPERSONAL);
+		    (void) printf ("\tSORTTMP = \"%s\"\n", SORTTMP);
+		    (void) printf ("\tSPELL_XREF = \"%s\"\n", SPELL_XREF);
 		    (void) printf ("\tSTATSUFFIX = \"%s\"\n", STATSUFFIX);
 		    (void) printf ("\tTEMPNAME = \"%s\"\n", TEMPNAME);
 		    (void) printf ("\tTERMLIB = \"%s\"\n", TERMLIB);
+#if TERM_MODE == CBREAK
+		    (void) printf ("\tTERM_MODE = CBREAK\n");
+#else /* TERM_MODE */
+		    (void) printf ("\tTERM_MODE = RAW\n");
+#endif /* TERM_MODE */
+		    (void) printf ("\tTEXSKIP1 = \"%s\"\n", TEXSKIP1);
+		    (void) printf ("\tTEXSKIP1VAR = \"%s\"\n", TEXSKIP1VAR);
+		    (void) printf ("\tTEXSKIP2 = \"%s\"\n", TEXSKIP2);
+		    (void) printf ("\tTEXSKIP2VAR = \"%s\"\n", TEXSKIP2VAR);
 		    (void) printf ("\tTEXSPECIAL = \"%s\"\n", TEXSPECIAL);
+		    (void) printf ("\tTIB_XREF = \"%s\"\n", TIB_XREF);
 #ifdef TRUNCATEBAK
 		    (void) printf ("\tTRUNCATEBAK\n");
 #else /* TRUNCATEBAK */
@@ -543,6 +593,7 @@ int main (argc, argv)
 		    (void) printf ("\t!USESH\n");
 #endif /* USESH */
 		    (void) printf ("\tWORDS = \"%s\"\n", WORDS);
+		    (void) printf ("\tW_OK = %d\n", W_OK);
 		    (void) printf ("\tYACC = \"%s\"\n", YACC);
 		    }
 		exit (0);
@@ -571,6 +622,14 @@ int main (argc, argv)
 		if (preftype == NULL)
 		    preftype = "sgml";
 		break;
+	    case 'o':                   /* Ordinary text mode */
+		if (arglen > 2)
+		    usage ();
+		tflag = DEFORMAT_NONE;
+		deftflag = DEFORMAT_NONE;
+		if (preftype == NULL)
+		    preftype = "plain";
+		break;
 	    case 'k':			/* Set keyword tables */
 		p = (*argv) + 2;
 		argv++;
@@ -653,7 +712,7 @@ int main (argc, argv)
 		eflag = 1;
 		if ((*argv)[2] == 'e')
 		    eflag = 2;
-		else if ((*argv)[2] >= '1'  &&  (*argv)[2] <= '4')
+		else if ((*argv)[2] >= '1'  &&  (*argv)[2] <= '5')
 		    eflag = (*argv)[2] - '0';
 		else if ((*argv)[2] != '\0')
 		    usage ();
@@ -774,10 +833,10 @@ int main (argc, argv)
 			usage ();
 		    p = *argv;
 		    }
-		if (index (p, '/') != NULL)
+		if (last_slash (p) != NULL)
 		    (void) strcpy (hashname, p);
 		else
-		    (void) sprintf (hashname, "%s/%s", LIBDIR, p);
+		    (void) sprintf (hashname, "%s/%s", libdir, p);
 		if (cpd == NULL  &&  *p != '\0')
 		    LibDict = p;
 		p = rindex (p, '.');
@@ -787,7 +846,7 @@ int main (argc, argv)
 		    (void) strcat (hashname, HASHSUFFIX);
 		if (LibDict != NULL)
 		    {
-		    p = rindex (LibDict, '/');
+		    p = last_slash (LibDict);
 		    if (p != NULL)
 			LibDict = p + 1;
 		    }
@@ -845,7 +904,7 @@ int main (argc, argv)
      */
     for (argno = 0;  argno < argc;  argno++)
 	{
-	if (access (argv[argno], 4) >= 0)
+	if (access (argv[argno], R_OK) >= 0)
 	    break;
 	}
     if (argno >= argc  &&  !lflag  &&  !aflag  &&  !eflag  &&  !dumpflag)
@@ -857,6 +916,9 @@ int main (argc, argv)
     if (linit () < 0)
 	exit (1);
 
+    if (preftype == NULL)
+	preftype = getenv (CHARSETVAR);
+
     if (preftype != NULL)
 	{
 	prefstringchar =
@@ -872,9 +934,9 @@ int main (argc, argv)
 	    }
 	}
     if (prefstringchar < 0)
-	defdupchar = 0;
+	defstringgroup = 0;
     else
-	defdupchar = prefstringchar;
+	defstringgroup = prefstringchar;
 
     if (compoundflag < 0)
 	compoundflag = hashheader.compoundflag;
@@ -931,12 +993,21 @@ int main (argc, argv)
 #endif /* __bsdi__ */
     if (lflag)
 	{
-	infile = setupdefmt(NULL);
+	infile = setupdefmt(NULL, NULL);
 	outfile = stdout;
 	checkfile ();
 	exit (0);
 	}
 
+    /*
+     * If there is a log directory, open a log file.  If the open
+     * fails, we just won't log.
+     */
+    (void) sprintf (logfilename, "%s/%s/%s",
+      getenv ("HOME") == NULL ? "" : getenv ("HOME"),
+      DEFLOGDIR, LibDict);
+    logfile = fopen (logfilename, "a");
+
     terminit ();
 
     while (argc--)
@@ -978,27 +1049,27 @@ static void dofile (filename)
 	}
     if (prefstringchar < 0)
 	{
-	defdupchar =
+	defstringgroup =
 	  findfiletype (filename, 0, deftflag < 0 ? &tflag : (int *) NULL);
-	if (defdupchar < 0)
-	    defdupchar = 0;
+	if (defstringgroup < 0)
+	    defstringgroup = 0;
 	}
 
-    if ((infile = setupdefmt (filename)) == NULL)
+    if ((infile = setupdefmt (filename, &statbuf)) == NULL)
 	{
-	(void) fprintf (stderr, CANT_OPEN, filename);
+	(void) fprintf (stderr, CANT_OPEN, filename, MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	return;
 	}
 
-    readonly = access (filename, 2) < 0;
+    readonly = access (filename, W_OK) < 0;
     if (readonly)
 	{
-	(void) fprintf (stderr, ISPELL_C_CANT_WRITE, filename);
+	(void) fprintf (stderr, ISPELL_C_CANT_WRITE, filename,
+	  MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	}
 
-    (void) fstat (fileno (infile), &statbuf);
     /*
      * Security notes: TEMPNAME must be less than MAXPATHLEN - 1.  If
      * the system has O_EXCL but not mkstemp, the temporary file will
@@ -1006,7 +1077,28 @@ static void dofile (filename)
      * unfortunately isn't available anywhere).  In other words, don't
      * worry about the security of this hunk of code.
      */
-    (void) strcpy (tempfile, TEMPNAME);
+    if (last_slash (TEMPNAME) != NULL)
+	(void) strcpy (tempfile, TEMPNAME);
+    else
+	{
+	char *tmp = getenv ("TMPDIR");
+	int   lastchar;
+
+	if (tmp == NULL)
+	    tmp = getenv ("TEMP");
+	if (tmp == NULL)
+	    tmp = getenv ("TMP");
+	if (tmp == NULL)
+#ifdef P_tmpdir
+	    tmp = P_tmpdir;
+#else
+	    tmp = "/tmp";
+#endif
+	lastchar = tmp[strlen (tmp) - 1];
+	(void) sprintf (tempfile, "%s%s%s", tmp,
+			IS_SLASH (lastchar) ? "" : "/",
+			TEMPNAME);
+	}
 #ifdef NO_MKSTEMP
     if (mktemp (tempfile) == NULL  ||  tempfile[0] == '\0'
 #ifdef O_EXCL
@@ -1022,11 +1114,20 @@ static void dofile (filename)
 	{
 	(void) fprintf (stderr, CANT_CREATE,
 	  (tempfile == NULL  ||  tempfile[0] == '\0')
-	    ? "temporary file" : tempfile);
+	    ? "temporary file" : tempfile,
+	  MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	return;
 	}
+#ifndef MSDOS
+    /*
+    ** This is usually a no-op on MS-DOS, but with file-sharing
+    ** installed, it was reported to produce empty spelled files!
+    ** Apparently, the file-sharing module would close the file when
+    ** `chmod' is called.
+    */
     (void) chmod (tempfile, statbuf.st_mode);
+#endif
 
     quit = 0;
     changes = 0;
@@ -1048,7 +1149,7 @@ static void dofile (filename)
  * Set up to externally deformat a file.
  *
  * If no deformatter is provided, we return either standard input (if
- * filename is NULL) or a the result of opening the specified file.
+ * filename is NULL) or the result of opening the specified file.
  *
  * If a deformatter is provided, but no filename is provided, we
  * assume that the input comes from stdin, and we set up the
@@ -1063,10 +1164,16 @@ static void dofile (filename)
  * so that it can read the unfiltered data directly from the input
  * file.
  *
+ * If "statbuf" is non-NULL, it will be filled in with the result of
+ * stat-ing the file.  If the stat fails, statbuf->st_mode will be set
+ * to a reasonable default value.  The contents of statbuf are
+ * undefined if the input file or the filter cannot be opened.
+ *
  * Note that external deformatters do not work with named pipes as input.
  */
-static FILE * setupdefmt (filename)
+static FILE * setupdefmt (filename, statbuf)
     char *		filename;	/* File to open, if non-NULL */
+    struct stat *	statbuf;	/* Buffer to hold file status */
     {
     FILE*		filteredfile;	/* Access to the filtered file */
     int			inputfd;	/* Fd for access to file to open */
@@ -1079,9 +1186,13 @@ static FILE * setupdefmt (filename)
 	 * There is no deformatter.  Return either stdin or an open file.
 	 */
 	if (filename == NULL)
-	    return stdin;
+	    filteredfile = stdin;
 	else
-	    return fopen (filename, "r");
+	    filteredfile = fopen (filename, "r");
+	if (statbuf != NULL  &&  filteredfile != NULL
+	  &&  fstat (fileno (filteredfile), statbuf) == -1)
+	    statbuf->st_mode = DEFAULT_FILE_MODE;
+	return filteredfile;
 	}
     else if (filename == NULL)
 	{
@@ -1089,6 +1200,8 @@ static FILE * setupdefmt (filename)
 	 * We are reading from standard input.  Switch over to a
 	 * filtered version of stdin.
 	 */
+	if (statbuf != NULL  &&  fstat (fileno (stdin), statbuf) == -1)
+	    statbuf->st_mode = DEFAULT_FILE_MODE;
 	return popen (defmtpgm, "r");
 	}
     else
@@ -1111,21 +1224,23 @@ static FILE * setupdefmt (filename)
 	sourcefile = fopen (filename, "r");
 	if (sourcefile == NULL)
 	    return NULL;
+	if (statbuf != NULL  &&  fstat (fileno (sourcefile), statbuf) == -1)
+	    statbuf->st_mode = DEFAULT_FILE_MODE;
 	savedstdin = dup (0);
-	close (0);
 	inputfd = open (filename, 0);
 	if (inputfd < 0)
 	    return NULL;		/* Failed to open the file */
-	else if (inputfd != 0)
+	else if (dup2 (inputfd, 0) != 0)
 	    {
-	    (void) fprintf (stderr, ISPELL_C_UNEXPECTED_FD, filename);
+	    (void) fprintf (stderr, ISPELL_C_UNEXPECTED_FD, filename,
+	      MAYBE_CR (stderr));
 	    exit (1);
 	    }
 	filteredfile = popen (defmtpgm, "r");
-	close (0);
-	if (dup (savedstdin) != 0)
+	if (dup2 (savedstdin, 0) != 0)
 	    {
-	    (void) fprintf (stderr, ISPELL_C_UNEXPECTED_FD, filename);
+	    (void) fprintf (stderr, ISPELL_C_UNEXPECTED_FD, filename,
+	      MAYBE_CR (stderr));
 	    exit (1);
 	    }
 	close (savedstdin);
@@ -1143,7 +1258,8 @@ static void update_file (filename, statbuf)
 
     if ((infile = fopen (tempfile, "r")) == NULL)
 	{
-	(void) fprintf (stderr, ISPELL_C_TEMP_DISAPPEARED, tempfile);
+	(void) fprintf (stderr, ISPELL_C_TEMP_DISAPPEARED, tempfile,
+	  MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	return;
 	}
@@ -1151,27 +1267,73 @@ static void update_file (filename, statbuf)
 #ifdef TRUNCATEBAK
     (void) strncpy (bakfile, filename, sizeof bakfile - 1);
     bakfile[sizeof bakfile - 1] = '\0';
-    if (strcmp(BAKEXT, filename + strlen(filename) - sizeof BAKEXT - 1) != 0)
-	{
-	pathtail = rindex (bakfile, '/');
-	if (pathtail == NULL)
-	    pathtail = bakfile;
-	else
-	    pathtail++;
-	if (strlen (pathtail) > MAXNAMLEN - sizeof BAKEXT - 1)
-	    pathtail[MAXNAMLEN - sizeof BAKEXT -1] = '\0';
-	(void) strcat (pathtail, BAKEXT);
-	}
-#else
+#else /* TRUNCATEBAK */
     (void) sprintf (bakfile, "%.*s%s", (int) (sizeof bakfile - sizeof BAKEXT),
       filename, BAKEXT);
-#endif
-
-    pathtail = rindex (bakfile, '/');
+#endif /* TRUNCATEBAK */
+    pathtail = last_slash (bakfile);
     if (pathtail == NULL)
 	pathtail = bakfile;
     else
 	pathtail++;
+#ifdef TRUNCATEBAK
+    if (strcmp(BAKEXT, filename + strlen(filename) - sizeof BAKEXT + 1) != 0)
+	{
+	if (strlen (pathtail) > MAXNAMLEN - sizeof BAKEXT + 1)
+	    pathtail[MAXNAMLEN - sizeof BAKEXT + 1] = '\0';
+	(void) strcat (pathtail, BAKEXT);
+	}
+#endif /* TRUNCATEBAK */
+
+
+#ifdef MSDOS
+    if (pathconf (filename, _PC_NAME_MAX) <= MAXBASENAMELEN + MAXEXTLEN + 1)
+	{
+	/*
+	** Excessive characters beyond 8+3 will be truncated by the
+	** OS.  Ensure the backup extension won't be truncated, and
+	** that we don't create an invalid filename (e.g., more than
+	** one dot).  Allow use of BAKEXT without a leading dot (such
+	** as "~").
+	*/
+	char *last_dot = rindex (pathtail, '.');
+
+	/*
+	** If no dot in backup filename, make BAKEXT be the extension.
+	** This ensures we don't truncate the name more than necessary.
+	*/
+	if (last_dot == NULL  &&  strlen (pathtail) > MAXBASENAMELEN)
+	    {
+	    pathtail[MAXBASENAMELEN] = '.';
+	    /*
+	    ** BAKEXT cannot include a dot here (or we would have
+	    ** found it above, and last_dot would not be NULL).
+	    */
+	    strcpy (pathtail + MAXBASENAMELEN + 1, BAKEXT);
+	    }
+	else if (last_dot != NULL)
+	    {
+	    char *p = pathtail;
+	    size_t ext_len = strlen (last_dot);
+
+	    /* Convert all dots but the last to underscores. */
+	    while (p < last_dot && *p)
+		{
+		if (*p == '.')
+		    *p = '_';
+		p++;
+		}
+
+	    /* Make sure we preserve as much of BAKEXT as we can. */
+	    if (ext_len > MAXEXTLEN && ext_len > sizeof (BAKEXT) - 1)
+		strcpy (MAXEXTLEN <= sizeof (BAKEXT) - 1
+		    ? last_dot + 1
+		    : last_dot + MAXEXTLEN - sizeof (BAKEXT) + 1,
+		  BAKEXT);
+	    }
+	}
+#endif /* MSDOS */
+
     if (strncmp (filename, bakfile, pathtail - bakfile + MAXNAMLEN) != 0)
 	(void) unlink (bakfile);	/* unlink so we can write a new one. */
 #ifdef HAS_RENAME
@@ -1184,12 +1346,20 @@ static void update_file (filename, statbuf)
     /* if we can't write new, preserve .bak regardless of xflag */
     if ((outfile = fopen (filename, "w")) == NULL)
 	{
-	(void) fprintf (stderr, CANT_CREATE, filename);
+	(void) fprintf (stderr, CANT_CREATE, filename, MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	return;
 	}
 
+#ifndef MSDOS
+    /*
+    ** This is usually a no-op on MS-DOS, but with file-sharing
+    ** installed, it was reported to produce empty spelled files!
+    ** Apparently, the file-sharing module would close the file when
+    ** `chmod' is called.
+    */
     (void) chmod (filename, statbuf->st_mode);
+#endif
 
     while ((c = getc (infile)) != EOF)
 	(void) putc (c, outfile);
@@ -1208,6 +1378,7 @@ static void expandmode (option)
 					/* 2 = original line + expansions */
 					/* 3 = original paired w/ expansions */
 					/* 4 = add length ratio */
+                                        /* 5 = root + flags used, expansion */
     {
     char		buf[BUFSIZ];
     int			explength;	/* Total length of all expansions */
@@ -1252,7 +1423,8 @@ static void expandmode (option)
 		if (temp >= 0  &&  temp <= LARGESTFLAG)
 		    SETMASKBIT (mask, temp);
 		else
-		    (void) fprintf (stderr, BAD_FLAG, (unsigned char) *flagp);
+		    (void) fprintf (stderr, BAD_FLAG, MAYBE_CR (stderr),
+		      (unsigned char) *flagp, MAYBE_CR (stderr));
 		flagp++;
 		/* Accept old-format dicts with extra slashes */
 		if (*flagp == hashheader.flagmarker)
@@ -1280,3 +1452,37 @@ static void expandmode (option)
 	(void) fflush (stdout);
 	}
     }
+
+/*
+** A trivial wrapper for rindex (file '/') on Unix, but
+** saves a lot of ifdef-ing on MS-DOS.
+*/
+char * last_slash (file)
+    char *		file;		/* String to search for / or \ */
+    {
+#ifdef MSDOS
+    char *		backslash;	/* Position of last backslash */
+#endif /* MSDOS */
+    char *		slash;		/* Position of last slash */
+
+    slash = rindex (file, '/');
+
+#ifdef MSDOS
+    /*
+    ** We can have both forward- and backslashes; find the rightmost
+    ** one of either type.
+    */
+    backslash = rindex (file, '\\');
+    if (slash == NULL  ||  (backslash != NULL  &&  backslash > slash))
+	slash = backslash;
+
+    /*
+    ** If there is no backslash, but the first two characters are a
+    ** letter and a colon, the basename begins right after them.
+    */
+    if (slash == NULL  &&  file[0] != '\0'  &&  file[1] == ':')
+	slash = file + 1;
+#endif
+
+    return slash;
+    }
