@@ -1,7 +1,7 @@
 #ifndef MSGS_H_INCLUDED
 #define MSGS_H_INCLUDED
 /*
- * $Id: msgs.h,v 1.46 2008/02/22 06:19:19 geoff Exp $
+ * $Id: msgs.h,v 1.48 2021-01-08 16:44:27-08 geoff Exp $
  *
  * Copyright 1992, 1993, 1999, 2001, Geoff Kuenning, Claremont, CA
  * All rights reserved.
@@ -65,7 +65,13 @@
 
 /*
  * $Log: msgs.h,v $
- * Revision 1.46  2008/02/22 06:19:19  geoff
+ * Revision 1.48  2021-01-08 16:44:27-08  geoff
+ * Conditionalize the help messages so that warnings in correct.c go away.
+ *
+ * Revision 1.47  2020-12-30 22:20:19-08  geoff
+ * Add the illegal-option message.
+ *
+ * Revision 1.46  2008-02-21 22:19:19-08  geoff
  * Improve the hash overflow message.
  *
  * Revision 1.45  2005/04/28 00:19:48  geoff
@@ -185,10 +191,12 @@
 #define CORR_C_HELP_4		"next to each one.  You have the option of replacing the word%s\n"
 #define CORR_C_HELP_5		"completely, or choosing one of the suggested words.%s\n"
     /* You may add HELP_6 through HELP_9 if your language needs more lines */
+#if 0
 #define CORR_C_HELP_6		""
 #define CORR_C_HELP_7		""
 #define CORR_C_HELP_8		""
 #define CORR_C_HELP_9		""
+#endif
 #define CORR_C_HELP_COMMANDS	"%s\nCommands are:%s\n%s\n"
 #define CORR_C_HELP_R_CMD	"R       Replace the misspelled word completely.%s\n"
 #define CORR_C_HELP_BLANK	"Space   Accept the word this time only.%s\n"
@@ -300,6 +308,7 @@
 #define PARSE_Y_LONG_FLAG	"Flag must be single character"
 #define PARSE_Y_BAD_FLAG	"Flag must be alphabetic"
 #define PARSE_Y_DUP_FLAG	"Duplicate flag"
+#define PARSE_Y_BAD_OPTION	"Illegal option"
 #define PARSE_Y_NO_SPACE	"Out of memory"
 #define PARSE_Y_NEED_BLANK	"Single characters must be separated by a blank"
 #define PARSE_Y_MANY_CONDS	"Too many conditions;  8 maximum"
