@@ -1,6 +1,6 @@
 #ifndef lint
 static char Rcs_Id[] =
-    "$Id: tree.c,v 1.63 2001/07/25 21:51:46 geoff Exp $";
+    "$Id: tree.c,v 1.66 2005/04/14 14:43:46 geoff Exp $";
 #endif
 
 /*
@@ -9,7 +9,8 @@ static char Rcs_Id[] =
  * Pace Willisson, 1983
  * Hash support added by Geoff Kuenning, 1987
  *
- * Copyright 1987-1989, 1992, 1993, 1999, 2001, Geoff Kuenning, Claremont, CA
+ * Copyright 1987-1989, 1992, 1993, 1999, 2001, 2005, Geoff Kuenning,
+ * Claremont, CA
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -25,16 +26,8 @@ static char Rcs_Id[] =
  *    such.  Binary redistributions based on modified source code
  *    must be clearly marked as modified versions in the documentation
  *    and/or other materials provided with the distribution.
- * 4. Any web site or other electronic service that offers ispell for
- *    download or other electronic transfer as a separate entity, in
- *    either source or binary form, must also include a prominent statement
- *    indicating that information about ispell can be obtained from the
- *    following Web site URL:
- *	http://fmg-www.cs.ucla.edu/geoff/ispell.html
- *    If the offering service supports hyperlinks, the aforementioned
- *    Web site must also be offered as a hyperlink.  Condition #4 does
- *    not apply if ispell is offered only as part of a larger, aggregated
- *    product such as a word processor or packaged operating system.
+ * 4. The code that causes the 'ispell -v' command to display a prominent
+ *    link to the official ispell Web site may not be removed.
  * 5. The name of Geoff Kuenning may not be used to endorse or promote
  *    products derived from this software without specific prior
  *    written permission.
@@ -54,6 +47,16 @@ static char Rcs_Id[] =
 
 /*
  * $Log: tree.c,v $
+ * Revision 1.66  2005/04/14 14:43:46  geoff
+ * Update copyright
+ *
+ * Revision 1.65  2005/04/14 14:38:23  geoff
+ * Update license.  Don't segfault if HOME is unset in the environment
+ * and PDICTHOME isn't set in config.X
+ *
+ * Revision 1.64  2001/09/06 00:30:29  geoff
+ * Many changes from Eli Zaretskii to support DJGPP compilation.
+ *
  * Revision 1.63  2001/07/25 21:51:46  geoff
  * Minor license update.
  *
@@ -213,7 +216,8 @@ void treeinit (p, LibDict)
 	    if (seconddict[0] != '\0')
 		(void) strcpy (personaldict, seconddict);
 	    else
-		(void) sprintf (personaldict, "%s/%s%s", h, DEFPDICT, LibDict);
+		(void) sprintf (personaldict, "%s/%s%s", h == NULL ? "" : h,
+		  DEFPDICT, LibDict);
 	    }
 	if (dictf != NULL)
 	    {
@@ -233,8 +237,16 @@ void treeinit (p, LibDict)
 	** with "./" and "../" is considered an absolute path, since this
 	** still means we can't prepend HOME.
 	*/
-	abspath = (*p == '/'  ||  strncmp (p, "./", 2) == 0
-	  ||  strncmp (p, "../", 3) == 0);
+	abspath = IS_SLASH (*p)  ||  strncmp (p, "./", 2) == 0
+	  ||  strncmp (p, "../", 3) == 0;
+#ifdef MSDOS
+	/*
+	** DTRT with drive-letter braindamage and with backslashes.
+	*/
+	abspath |= strncmp (p, ".\\", 2) == 0
+	  || strncmp (p, "..\\", 3) == 0
+	  || (p[0] != '\0'  &&  p[1] == ':');
+#endif
 	if (abspath)
 	    {
 	    (void) strcpy (personaldict, p);
@@ -276,7 +288,7 @@ void treeinit (p, LibDict)
 	     */
 	    if (dictf == NULL)
 		{
-		(void) fprintf (stderr, CANT_OPEN, p);
+		(void) fprintf (stderr, CANT_OPEN, p, MAYBE_CR (stderr));
 		perror ("");
 		return;
 		}
@@ -284,9 +296,10 @@ void treeinit (p, LibDict)
 	}
 
     if (!lflag  &&  !aflag
-      &&  access (personaldict, 2) < 0  &&  errno != ENOENT)
+      &&  access (personaldict, W_OK) < 0  &&  errno != ENOENT)
 	{
-	(void) fprintf (stderr, TREE_C_CANT_UPDATE, personaldict);
+	(void) fprintf (stderr, TREE_C_CANT_UPDATE, personaldict,
+	  MAYBE_CR (stderr));
 	(void) sleep ((unsigned) 2);
 	}
     }
@@ -359,7 +372,7 @@ void treeinsert (word, wordlen, keep)
 	  (struct dent *) calloc ((unsigned) pershsize, sizeof (struct dent));
 	if (pershtab == NULL)
 	    {
-	    (void) fprintf (stderr, TREE_C_NO_SPACE);
+	    (void) fprintf (stderr, TREE_C_NO_SPACE, MAYBE_CR (stderr));
 	    /*
 	     * Try to continue anyway, since our overflow
 	     * algorithm can handle an overfull (100%+) table,
@@ -369,7 +382,7 @@ void treeinsert (word, wordlen, keep)
 	     */
 	    if (oldhtab == NULL)
 		exit (1);		/* No old table, can't go on */
-	    (void) fprintf (stderr, TREE_C_TRY_ANYWAY);
+	    (void) fprintf (stderr, TREE_C_TRY_ANYWAY, MAYBE_CR (stderr));
 	    cantexpand = 1;		/* Suppress further messages */
 	    pershsize = oldhsize;	/* Put things back */
 	    pershtab = oldhtab;		/* ... */
@@ -471,7 +484,7 @@ static struct dent * tinsert (proto)
 	hp = (struct dent *) calloc (1, sizeof (struct dent));
 	if (hp == NULL)
 	    {
-	    (void) fprintf (stderr, TREE_C_NO_SPACE);
+	    (void) fprintf (stderr, TREE_C_NO_SPACE, MAYBE_CR (stderr));
 	    exit (1);
 	    }
 	}
@@ -540,7 +553,7 @@ void treeoutput ()
 
     if ((dictf = fopen (personaldict, "w")) == NULL)
 	{
-	(void) fprintf (stderr, CANT_CREATE, personaldict);
+	(void) fprintf (stderr, CANT_CREATE, personaldict, MAYBE_CR (stderr));
 	return;
 	}
 
@@ -690,7 +703,7 @@ char * do_regex_lookup (expr, whence)
     static int		    curindex;
     static struct dent *    curpersent;
     static int		    curpersindex;
-    static char *	    cmp_expr;
+    static REGCTYPE	    cmp_expr = (REGCTYPE) NULL;
     char		    dummy[INPUTWORDLEN + MAXAFFIXLEN];
     ichar_t *		    is;
 
@@ -699,7 +712,8 @@ char * do_regex_lookup (expr, whence)
 	is = strtosichar (expr, 0);
 	upcase (is);
 	expr = ichartosstr (is, 1);
-        cmp_expr = REGCMP (expr);
+	REGFREE (cmp_expr);	/* free previous compiled pattern, if any */
+        cmp_expr = REGCMP (cmp_expr, expr);
         curent = hashtbl;
         curindex = 0;
         curpersent = pershtab;
