@@ -10,10 +10,10 @@
  */
 
 static char * Version_ID[] = {
-    "@(#) International Ispell Version 3.4.00 8 Feb 2015",
+    "@(#) International Ispell Version 3.4.02 08 Jan 2021",
     "@(#) Copyright (c), 1983, by Pace Willisson",
     "@(#) International version Copyright (c) 1987, 1988, 1990-1995, 1999,",
-    "@(#) 2001, 2002, 2005, 2015 by Geoff Kuenning, Claremont, CA.",
+    "@(#) 2001, 2002, 2005, 2015, 2021 by Geoff Kuenning, Claremont, CA.",
     "@(#) All rights reserved.",
     "@(#)",
     "@(#) The official ispell Web site is at:",
@@ -63,6 +63,6 @@ static char * Version_ID[] = {
 };
 
 static char RCS_Version_ID[] =
-    "$Id: version.h,v 1.60 2015-02-08 16:44:37-08 geoff Exp $";
+    "$Id: version.h,v 1.62 2021-01-09 13:12:35-08 geoff Exp $";
 
 #endif /* VERSION_H_INCLUDED */
